package kcode.statexmi2java.codegenerator;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.RuntimeErrorException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import kcode.statexmi2java.codegenerator.templates.ITemplator;
import kcode.statexmi2java.smrepresentation.Behavior;
import kcode.statexmi2java.smrepresentation.State;
import kcode.statexmi2java.smrepresentation.StateMachine;
import kcode.statexmi2java.smrepresentation.Transition;

/**
 * Allows for generation of java sourcecode in a target folder from a statemachine
 */
public class JavaGenerator implements CodeGenerator {

  protected final Logger logger = LoggerFactory.getLogger(JavaGenerator.class);

  // dependencies
  protected ITemplator   templator;
  protected Class<?>     loaderClass;
  protected String       classLoaderResourcePrefix;

  /**
   * @param templator templator to be used
   * @param loaderClass class used to load resources
   * @param templateSubDir subdir for resources
   */
  public JavaGenerator(ITemplator templator, Class<?> loaderClass, String templateSubDir) {
    this.templator = templator;
    this.loaderClass = loaderClass;
    this.classLoaderResourcePrefix = templateSubDir;
  }

  // TODO [REFACTOR] generateCodeFor(SM) method
  /**
   * generates java sourcecode in target folder from a statemachine
   * 
   * @param stateMachine statemachine representation
   * @param toFolder target folder to generate code in
   * @throws IOException thrown when problems with target folder or files occur
   */
  @Override
  public void generateCodeFor(final StateMachine stateMachine, final File toFolder) throws IOException {

    // create target dir (and path to it) if not exists yet
    final String targetDir = toFolder + File.separator + "codetemplates";
    final File targetDirFile = new File(targetDir);
    if (!targetDirFile.exists()) {
      if (!targetDirFile.mkdirs()) {
        logger.error("Code Target-Sub-Directory {} could not be created.", targetDirFile.toString());
      }
    }

    copyCommonTemplateFiles(targetDir);

    // general transition
    this.write(templator.getTransitionCode("StateTransitionConcrete", null, null), toFolder + File.separator
        + "StateTransitionConcrete.java");

    // individual transitions
    final Map<String, Integer> nameIterator = new HashMap<String, Integer>();
    final List<String> transitionInstantiationCode = new ArrayList<String>();
    final List<String> triggers = new ArrayList<String>();
    final List<String> stateClassNames = new ArrayList<String>();
    final List<State> transitionsGenerated = new ArrayList<State>();
    String initialStateID = null;
    for (Transition transition : stateMachine.transitions) {
      // only parse transition if it is a valid transition (has a source and target state)
      if (transition.from == null || transition.to == null) {
        logger.warn("Skipping invalid transition " + transition.name);
      } else {
        // transition name
        String name = "Transition" + transition.from.name + transition.to.name;
        Integer nr = nameIterator.get(name);
        if (nr == null) {
          // name was not used previously, just put a number into the map
          nameIterator.put(name, 1);
        } else {
          // name was used previously, increment the counter and append it to the name
          nr++;
          nameIterator.put(name, nr);
          name += nr;
        }
        // transition guard code
        final String effectCode = transition.effect != null ? "@Override\nprotected void executeEffect(Object context){\n  //TODO [gen] "
            + transition.effect.name + " " + transition.effect.body + "\n}\n" : null;
        this.write(templator.getTransitionCode(name, transition.guard.isEmpty() ? null : transition.guard, effectCode),
            toFolder + File.separator + name + ".java");

        // states
        if (!transitionsGenerated.contains(transition.from)) {
          initialStateID = transitionStateGen(transition.from, toFolder, stateClassNames, initialStateID);
          transitionsGenerated.add(transition.from);
        }
        if (!transitionsGenerated.contains(transition.to)) {
          initialStateID = transitionStateGen(transition.to, toFolder, stateClassNames, initialStateID);
          transitionsGenerated.add(transition.to);
        }

        String triggerEnumName;
        if (transition.triggers.size() == 0) {
          Logger logger = LoggerFactory.getLogger(this.getClass());
          logger.warn("Transition " + name + " has no trigger(s) associated!");

          triggerEnumName = "Trigger.ALWAYS";
        } else {
          if (transition.triggers.size() > 1) {
            Logger logger = LoggerFactory.getLogger(this.getClass());
            logger.warn("Transition " + name + " has more than 1 triggers associated!");
          }

          triggerEnumName = "Trigger." + transition.triggers.get(0).toUpperCase();

          final StringBuilder sB = new StringBuilder();
          sB.append("ObjectState<StateID, Trigger> stFrom = new " + transition.from.name + "(StateID."
              + transition.from.name.toUpperCase() + ");\n");
          sB.append("ObjectState<StateID, Trigger> stTo = new " + transition.to.name + "(StateID."
              + transition.to.name.toUpperCase() + ");\n");
          sB.append("t = new " + name + "<StateID, Trigger>(stFrom, stTo, " + triggerEnumName + ");\n");
          sB.append("stFrom.addTransition(t);\n");
          sB.append("states.put(stFrom.getStateID(), stFrom);\n");
          sB.append("stTo.addTransition(t);\n");
          sB.append("states.put(stTo.getStateID(), stTo);\n");
          transitionInstantiationCode.add(sB.toString());

          // accumulate triggers
          for (String trigger : transition.triggers) {
            if (!triggers.contains(trigger)) {
              triggers.add(trigger);
            }
          }
        }
      }
    }

    // generate triggers file
    this.write(templator.getTriggerEnumCode(triggers), toFolder + File.separator + "Trigger.java");

    // generate stateIDs file
    this.write(templator.getStateIDEnumCode(stateClassNames), toFolder + File.separator + "StateID.java");

    // factory
    final StringBuilder createStatesCode = new StringBuilder();
    for (String stateClassName : stateClassNames) {
      createStatesCode.append("states.put(StateID.").append(stateClassName.toUpperCase()).append(", new ")
          .append(stateClassName).append("(StateID.").append(stateClassName.toUpperCase()).append("));\n");
    }
    final StringBuilder createTransitionsCode = new StringBuilder("StateTransition<StateID, Trigger> t;\n");
    for (String initializationCode : transitionInstantiationCode) {
      createTransitionsCode.append(initializationCode).append("\n");
    }
    this.write(
        templator.getStateFactoryCode(createStatesCode.toString(), createTransitionsCode.toString(), initialStateID),
        toFolder + File.separator + "StateFactoryConcrete.java");
  }

  protected void copyCommonTemplateFiles(String targetDir) {
    try {
      // copy common templates to target folder
      final String resCommonPrefix = classLoaderResourcePrefix + "/codetemplates/";
      final String[] commonFilenames = { "IStateFactory.java", "ObjectState.java", "StateFactory.java",
          "StateTransition.java" };
      InputStream in;
      OutputStream out;
      byte[] buf = new byte[1024];
      int len;
      for (String fileName : commonFilenames) {
        in = loaderClass.getResourceAsStream(resCommonPrefix + fileName);
        if (in == null) {
          logger.error("Could not find Template file! Path: {}", resCommonPrefix);
          throw new RuntimeErrorException(new Error("Could not find mandatory Template file!"));
        }
        final File targetFile = new File(targetDir + File.separator + fileName);
        out = new FileOutputStream(targetFile);
        // Transfer bytes from in to out
        while ((len = in.read(buf)) > 0) {
          out.write(buf, 0, len);
        }
        in.close();
        out.close();
      }
    } catch (FileNotFoundException e) {
      e.printStackTrace();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  /**
   * @param state representational state
   * @param toFolder target folder to generate the transitioncode in
   * @param stateClassNames list of class names being used
   * @param initialStateID current ID for initial state, or null
   * @return initialStateID or this states ID if this state is an initial state
   */
  protected String transitionStateGen(State state, File toFolder, List<String> stateClassNames, String initialStateID)
      throws IOException {
    state.name = "State" + state.name;
    this.write(templator.getStateCode(state.name, getStateMethodsCode(state)), toFolder + File.separator + state.name
        + ".java");
    if (!stateClassNames.contains(state.name)) {
      stateClassNames.add(state.name);
    }
    if (state.isInitial && initialStateID == null) {
      initialStateID = state.name.toUpperCase();
    }
    return initialStateID;
  }

  /**
   * Get the code for a states methods (like guard)
   * 
   * @param sF representational state
   * @return Code for methods, or empty if no methods to be generated
   */
  protected String getStateMethodsCode(State sF) {
    return getStateMethodsCode(sF.entry, "onStateEntry") + getStateMethodsCode(sF.doActivity, "onStateDoActivity")
        + getStateMethodsCode(sF.exit, "onStateExit");
  }

  /**
   * Get the code for a states methods (like guard)
   * 
   * @param behavior representational behavior to generate methodcode for
   * @param methodname the name of the method to generate
   * @return method code
   */
  protected String getStateMethodsCode(Behavior behavior, String methodname) {
    return behavior != null ? "@Override\nprotected void " + methodname + "(Object context){\n  //TODO [gen] "
        + behavior.name + "\n" + behavior.body + "\n}\n" : "";
  }

  /**
   * Write content to a file
   * 
   * @param content the content to be written to a file
   * @param toFile the path to the file
   * @throws IOException
   */
  protected void write(String content, String toFile) throws IOException {
    this.write(content, new File(toFile));
  }

  /**
   * Write content to a file
   * 
   * @param content the content to be written to a file
   * @param toFile the File to write to
   * @throws IOException
   */
  protected void write(String content, File toFile) throws IOException {
    FileWriter fw;
    fw = new FileWriter(toFile);
    fw.append(content);
    fw.close();
  }

  /**
   * Copy the content of one file to another
   * 
   * @param src source File to read the content from
   * @param dst target File to copy the content to
   * @throws IOException
   */
  protected static void copy(File src, File dst) throws IOException {
    InputStream in = new FileInputStream(src);
    OutputStream out = new FileOutputStream(dst);

    // Transfer bytes from in to out
    byte[] buf = new byte[1024];
    int len;
    while ((len = in.read(buf)) > 0) {
      out.write(buf, 0, len);
    }
    in.close();
    out.close();
  }
}
