package kcode.statexmi2java.codegenerator.templates;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import freemarker.template.Configuration;
import freemarker.template.DefaultObjectWrapper;
import freemarker.template.Template;
import freemarker.template.TemplateException;

/**
 * Templator providing methods for getting code from templates, via FreeMarker library and templates
 */
public class TemplatorFreeMarker implements ITemplator {

  protected final Logger  logger  = LoggerFactory.getLogger(TemplatorFreeMarker.class);

  /**
   * FreeMarker Configuration used for processing templates
   */
  protected Configuration tmplCfg = new Configuration();

  /**
   * Instantiate Templator providing methods for getting code from templates, via FreeMarker library and templates
   * 
   * @param loaderClass class whichs classLoader to get resources from ( – runtime class object)
   * @param templateSubDir prefix for subdir on loaderClasses getResource
   */
  public TemplatorFreeMarker(Class<?> loaderClass, String templateSubDir) {

    // initialize the configuration
    tmplCfg.setClassForTemplateLoading(loaderClass, templateSubDir);
    // how templates will see the data-model
    tmplCfg.setObjectWrapper(new DefaultObjectWrapper());
  }

  /**
   * @see kcode.statexmi2java.codegenerator.templates.ITemplator#getTransitionCode(java.lang.String, java.lang.String, java.lang.String)
   */
  @Override
  public String getTransitionCode(String classname, String guard, String effectCode) {

    try {
      // template file
      Template tmpl = tmplCfg.getTemplate("transitionfromto.ftl");
      // data-model and vars
      Map<Object, Object> tmplData = new HashMap<Object, Object>();
      tmplData.put("classname", classname);
      tmplData.put("extendsclassname", "StateTransition");
      String classinner = "";
      if (guard != null) {
        classinner += getTransitionGuardCode(guard);
      }
      if (effectCode != null) {
        classinner += effectCode;
      }
      tmplData.put("classinner", classinner);
      // process template
      return this.processTemplate(tmpl, tmplData);

    } catch (IOException exc) {
      logger.error("could not find template");
      logger.debug(Arrays.toString(exc.getStackTrace()));
    }

    return null;
  }

  /**
   * get code for transition guard
   * 
   * @param guard the guard
   * @return code for the (transition) guard
   */
  public String getTransitionGuardCode(String guard) {
    // a guard passed -> process normally
    if (guard != null && !guard.isEmpty()) {
      try {
        // template file
        final Template tmpl = tmplCfg.getTemplate("transitionfromtoguard.ftl");
        // data-model
        final Map<Object, Object> tmplData = new HashMap<Object, Object>();
        tmplData.put("guard", guard);
        // process; get code
        return this.processTemplate(tmpl, tmplData);
      } catch (IOException exc) {
        logger.error("Could not find template");
        logger.debug(Arrays.toString(exc.getStackTrace()));
      }
    }
    return "";
  }

  /**
   * @see kcode.statexmi2java.codegenerator.templates.ITemplator#getStateCode(java.lang.String)
   */
  @Override
  public String getStateCode(String stateName) {
    return getStateCode(stateName, null);
  }

  /**
   * @see kcode.statexmi2java.codegenerator.templates.ITemplator#getStateCode(java.lang.String, java.lang.String)
   */
  @Override
  public String getStateCode(String stateName, String classinner) {
    try {
      Template tmpl = tmplCfg.getTemplate("state.ftl");

      /* Create data-model */
      Map<Object, Object> tmplData = new HashMap<Object, Object>();
      tmplData.put("classname", stateName);
      tmplData.put("classinner", classinner);

      return this.processTemplate(tmpl, tmplData);

    } catch (IOException exc) {
      logger.error("could not find template");
      logger.debug(Arrays.toString(exc.getStackTrace()));
    }

    return null;
  }

  /**
   * @see kcode.statexmi2java.codegenerator.templates.ITemplator#getTriggerEnumCode(java.util.List)
   */
  @Override
  public String getTriggerEnumCode(List<String> triggers) {
    try {
      // template file
      Template tmpl = tmplCfg.getTemplate("trigger.ftl");

      // concatenate the triggers to a code string
      StringBuilder vals = new StringBuilder();
      for (int i = 0; i < triggers.size(); i++) {
        if (i != 0) {
          vals.append(", ");
        }
        vals.append(triggers.get(i).replaceAll(" ", "").toUpperCase());
      }

      // data-model
      final Map<Object, Object> tmplData = new HashMap<Object, Object>();
      tmplData.put("enumvalues", vals.toString());
      // process
      return this.processTemplate(tmpl, tmplData);

    } catch (IOException exc) {
      logger.error("could not find template");
      logger.debug(Arrays.toString(exc.getStackTrace()));
    }
    return null;
  }

  /**
   * @see kcode.statexmi2java.codegenerator.templates.ITemplator#getStateIDEnumCode(java.util.List)
   */
  @Override
  public String getStateIDEnumCode(List<String> stateClassNames) {
    try {
      // template file
      Template tmpl = tmplCfg.getTemplate("stateid.ftl");
      // data-model
      StringBuilder vals = new StringBuilder();
      for (int i = 0; i < stateClassNames.size(); i++) {
        switch (i) {
          case 0:
            vals.append(stateClassNames.get(i).toUpperCase());
            break;
          default:
            vals.append(", " + stateClassNames.get(i).toUpperCase());
        }
      }

      Map<Object, Object> tmplData = new HashMap<Object, Object>();
      tmplData.put("enumvalues", vals.toString());
      // process
      return this.processTemplate(tmpl, tmplData);

    } catch (IOException exc) {
      logger.error("could not find template");
      logger.debug(Arrays.toString(exc.getStackTrace()));
    }
    return null;
  }

  /**
   * @see kcode.statexmi2java.codegenerator.templates.ITemplator#getStateFactoryCode(java.lang.String, java.lang.String, java.lang.String)
   */
  @Override
  public String getStateFactoryCode(String createStatesCode, String createTransitionsCode, String initialStateID) {
    try {
      Template tmpl = tmplCfg.getTemplate("statefactory.ftl");

      // data-model
      Map<Object, Object> tmplData = new HashMap<Object, Object>();
      tmplData.put("classname", "StateFactoryConcrete");
      //TODO [REFACTOR] remove; not used anymore
      tmplData.put("createstates", createStatesCode);
      tmplData.put("createstatestransitions", createTransitionsCode);
      tmplData.put("initialstateid", initialStateID);

      // process the template
      return this.processTemplate(tmpl, tmplData);

    } catch (IOException exc) {
      logger.error("could not find template");
      logger.debug(Arrays.toString(exc.getStackTrace()));
    }
    return null;
  }

  /**
   * Process the given template with the given data and return the processed template as string
   * Wraps exceptions
   * @param tmpl template
   * @param data data map
   * @return processed template
   */
  protected String processTemplate(Template tmpl, Map<Object, Object> data) {
    Writer outStrWr = new StringWriter();
    try {
      // process and return result
      tmpl.process(data, outStrWr);
      return outStrWr.toString();
    } catch (TemplateException e) {
      logger.error("TemplateException processing template");
      logger.debug(Arrays.toString(e.getStackTrace()));
    } catch (IOException exc) {
      logger.error("IOException processing template");
      logger.debug(Arrays.toString(exc.getStackTrace()));
    }
    return null;
  }
}
