package kcode.statexmi2java.gui;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.List;

import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.LayoutStyle;
import javax.swing.filechooser.FileFilter;

import kcode.statexmi2java.xmireader.IXMIReader;

import org.ciscavate.cjwizard.WizardPage;
import org.ciscavate.cjwizard.WizardSettings;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * CJWizard wizard page for browsing the source XMI file and target folder. On page-finish the XMI file is parsed with the constructor-passed IXMIReader and the
 * settings updated with the statemachines.
 */
public class WizardPageBrowse extends WizardPage {
  private static final long serialVersionUID = 1L;
  // text
  private static final String TEXT_BROWSE = "Browse";
  private static final String TEXT_LABEL_TARGETFOLDER = "Zielverzeichnis:";
  // filechooser file filter constants
  protected static final String FILECHOOSER_FILEENDING_DESC = "XMI (*.xmi)";
  private static final String FILECHOOSER_FILEENDING_XML = ".xml";
  protected static final String FILECHOOSER_FILEENDING_XMI = ".xmi";
  // page constants
  public static final String WIZARDPAGE_TITLE = TEXT_BROWSE;
  public static final String WIZARDPAGE_DESC  = "Browse Source XMI and Target Folder";

  protected final Logger           logger           = LoggerFactory.getLogger(WizardPageFactory.class);

  protected WizardPageBrowse instance;
  // dependencies
  protected IXMIReader       xmiReader;

  // UI items
  protected JLabel           xmiFilePathLabel;
  protected JLabel           targetFolderPathLabel;
  protected JTextField       xmiFilePath;
  protected JTextField       targetFolderPath;
  protected JButton          xmiFilePathBrowse;
  protected JButton          targetFolderPathBrowse;
  protected JPanel           mainPanel;

  /**
   * @param xmiReader mandatory
   */
  public WizardPageBrowse(IXMIReader xmiReader) {
    super(WIZARDPAGE_TITLE, WIZARDPAGE_DESC);
    this.xmiReader = xmiReader;
    initComponents();

    xmiFilePath.setText(Wizard.getPersistentPathForXMI());
    targetFolderPath.setText(Wizard.getPersistentPathForGen());

    instance = this;
  }

  @Override
  public void rendering(List<WizardPage> path, final WizardSettings settings) {
    super.rendering(path, settings);
    setFinishEnabled(false);
    checkEnableNextButton();
  }

  /*
   * @see org.ciscavate.cjwizard.WizardPage#updateSettings(org.ciscavate.cjwizard.WizardSettings)
   */
  @Override
  public void updateSettings(WizardSettings settings) {
    // save form input (source xmi filepath, target folder for generated source code)
    String xmiPath = xmiFilePath.getText();
    settings.put(xmiFilePath.getName(), xmiPath);
    settings.put(targetFolderPath.getName(), targetFolderPath.getText());

    // parse xmi file and save statemachine representations
    xmiReader.initialize(xmiPath);
    settings.put(Wizard.SETT_KEY_STATEMACHINES, xmiReader.getStateMachines());
    logger.debug("[Wizard:Browse:updateSettings] number of statemachines parsed: {}", xmiReader.getStateMachines()
        .size());
  }

  /**
   * Check whether the next button should be enabled or disabled, and do so
   */
  protected void checkEnableNextButton() {
    logger.debug("checkEnableNextButton: xmipath: \"{}\", genpath: \"{}\"", xmiFilePath.getText(),
        targetFolderPath.getText());
    if (xmiFilePath.getText().isEmpty() || targetFolderPath.getText().isEmpty()) {
      setNextEnabled(false);
    } else {
      setNextEnabled(true);
    }
  }

  protected void initComponents() {

    mainPanel = new JPanel();

    xmiFilePathLabel = new JLabel("XMI Filepath:");
    xmiFilePath = new JTextField();
    xmiFilePath.setName(Wizard.SETT_KEY_XMI);
    xmiFilePathBrowse = new JButton(TEXT_BROWSE);
    xmiFilePathBrowse.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        JFileChooser fileChooser = new JFileChooser();
        fileChooser.setAcceptAllFileFilterUsed(false);
        fileChooser.setSelectedFile(new File(xmiFilePath.getText()));
        fileChooser.setFileFilter(new FileFilter() {
          public boolean accept(File f) {
            String fn = f.getName().toLowerCase();
            return f.isDirectory() || fn.endsWith(FILECHOOSER_FILEENDING_XMI) || fn.endsWith(FILECHOOSER_FILEENDING_XML);
          }

          public String getDescription() {
            return FILECHOOSER_FILEENDING_DESC;
          }
        });

        if (fileChooser.showOpenDialog((Component) e.getSource()) == JFileChooser.APPROVE_OPTION) {
          xmiFilePath.setText(fileChooser.getSelectedFile().getAbsolutePath());
          instance.checkEnableNextButton();
        }
      }
    });

    targetFolderPathLabel = new JLabel(TEXT_LABEL_TARGETFOLDER);
    targetFolderPath = new JTextField();
    targetFolderPath.setName(Wizard.SETT_KEY_GENTARGET);
    targetFolderPathBrowse = new JButton(TEXT_BROWSE);
    targetFolderPathBrowse.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        JFileChooser fileChooser = new JFileChooser();
        fileChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
        fileChooser.setSelectedFile(new File(targetFolderPath.getText()));

        // if a directory is selected, adjust the textfield and check for enable/disable of next-button
        if (fileChooser.showSaveDialog((Component) e.getSource()) == JFileChooser.APPROVE_OPTION) {
          targetFolderPath.setText(fileChooser.getSelectedFile().getAbsolutePath());
          instance.checkEnableNextButton();
        }
      }
    });

    //TODO [REFACTOR] simplify GUI Layout code !?
    GroupLayout mainPanelLayout = new GroupLayout(mainPanel);
    mainPanelLayout.setHorizontalGroup(mainPanelLayout.createParallelGroup(GroupLayout.Alignment.LEADING).addGroup(
        mainPanelLayout
            .createSequentialGroup()
            .addGroup(
                mainPanelLayout
                    .createParallelGroup(GroupLayout.Alignment.LEADING)
                    .addComponent(targetFolderPathLabel)
                    .addGroup(
                        GroupLayout.Alignment.TRAILING,
                        mainPanelLayout
                            .createSequentialGroup()
                            .addGap(10, 10, 10)
                            .addGroup(
                                mainPanelLayout.createParallelGroup(GroupLayout.Alignment.TRAILING)
                                    .addComponent(xmiFilePath, GroupLayout.DEFAULT_SIZE, 379, Short.MAX_VALUE)
                                    .addComponent(targetFolderPath, GroupLayout.DEFAULT_SIZE, 379, Short.MAX_VALUE))
                            .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
                            .addGroup(
                                mainPanelLayout.createParallelGroup(GroupLayout.Alignment.LEADING)
                                    .addComponent(targetFolderPathBrowse).addComponent(xmiFilePathBrowse)))
                    .addComponent(xmiFilePathLabel)).addContainerGap()));
    mainPanelLayout.setVerticalGroup(mainPanelLayout.createParallelGroup(GroupLayout.Alignment.LEADING).addGroup(
        mainPanelLayout
            .createSequentialGroup()
            .addContainerGap()
            .addComponent(xmiFilePathLabel)
            .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
            .addGroup(
                mainPanelLayout
                    .createParallelGroup(GroupLayout.Alignment.BASELINE)
                    .addComponent(xmiFilePath, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE,
                        GroupLayout.PREFERRED_SIZE).addComponent(xmiFilePathBrowse))
            .addGap(18, 18, 18)
            .addComponent(targetFolderPathLabel)
            .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED)
            .addGroup(
                mainPanelLayout
                    .createParallelGroup(GroupLayout.Alignment.BASELINE)
                    .addComponent(targetFolderPath, GroupLayout.PREFERRED_SIZE, GroupLayout.DEFAULT_SIZE,
                        GroupLayout.PREFERRED_SIZE).addComponent(targetFolderPathBrowse))
            .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED, 79, Short.MAX_VALUE)
            .addGroup(mainPanelLayout.createParallelGroup(GroupLayout.Alignment.BASELINE)).addContainerGap()));
    
    mainPanel.setLayout(mainPanelLayout);
    add(mainPanel);
  }
}
