package kcode.statexmi2java.gui;

import java.io.File;
import java.io.IOException;
import java.util.List;

import kcode.statexmi2java.codegenerator.CodeGenerator;
import kcode.statexmi2java.smrepresentation.StateMachine;
import kcode.statexmi2java.xmireader.IXMIReader;

import org.ciscavate.cjwizard.PageFactory;
import org.ciscavate.cjwizard.WizardPage;
import org.ciscavate.cjwizard.WizardSettings;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * CJWizard wizard factory, navigating wizard pages
 */
public class WizardPageFactory implements PageFactory {

  protected final Logger  logger = LoggerFactory.getLogger(WizardPageFactory.class);

  protected IXMIReader    xmiReader;
  protected CodeGenerator codeGen;

  public WizardPageFactory(IXMIReader xmiReader, CodeGenerator codeGen) {
    super();
    this.xmiReader = xmiReader;
    this.codeGen = codeGen;
  }

  /**
   * create a wizard page for a path with specific settings
   * 
   * @param path List of page-names user has already visited (not current -> going back and forth will not change this)
   * @param settings Map-like (is not!) storage of settings
   * @return The next WizardPage.
   */
  @Override
  public WizardPage createPage(List<WizardPage> path, WizardSettings settings) {
    logger.debug("createPage called with: path {}", path);

    WizardPage page = null;

    if (path.size() == 0) {
      // initial wizard creation -> show first page
      page = new WizardPageBrowse(xmiReader);
    } else {
      // non-initial
      String lastPageTitle = path.get(path.size() - 1).getTitle();
      if (lastPageTitle.equals(WizardPageBrowse.WIZARDPAGE_TITLE)) {
        // last wizard page was browse-page -> parse state machines
        assert settings.get(Wizard.SETT_KEY_STATEMACHINES) instanceof List<?> : "Assertion failed: "
            + Wizard.SETT_KEY_STATEMACHINES + " is " + settings.get(Wizard.SETT_KEY_STATEMACHINES);
        @SuppressWarnings("unchecked")
        List<StateMachine> sMs = (List<StateMachine>) settings.get(Wizard.SETT_KEY_STATEMACHINES);
        int numberOfSMs = sMs.size();
        logger.debug("Number of statemachines: {}", numberOfSMs);

        if (numberOfSMs == 0) {
          // no statemachines found when parsing XMI
          page = createWizardPageError("No StateMachines were found in the imported XMI file.");
        } else if (numberOfSMs == 1) {
          // one statemachines found, thus generate the code and show the finish page
          try {
            File genTarget = new File((String) settings.get(Wizard.SETT_KEY_GENTARGET));
            generateCodeWith(codeGen, sMs.get(0), genTarget);
            page = createWizardPageFinish(settings);
          } catch (IOException e) {
            page = createWizardPageError("IO-Fehler: " + e.getMessage());
          }
        } else {
          // more than one statemachine was found -> show sm-selection page so user can select a statemachine to export
          page = new WizardPageDiagramSel(settings);
        }
      } else if (lastPageTitle.equals(WizardPageDiagramSel.WIZARDPAGE_TITLE)) {
        // last page was sm-selection -> generate code and display finish page
        try {
          @SuppressWarnings("unchecked")
          List<StateMachine> stMs = (List<StateMachine>) settings.get(Wizard.SETT_KEY_STATEMACHINES);
          Integer index = (Integer) settings.get(Wizard.SETT_KEY_DIAGRAMSEL);
          File genTarget = new File((String) settings.get(Wizard.SETT_KEY_GENTARGET));
          generateCodeWith(codeGen, stMs.get(index), genTarget);
          page = createWizardPageFinish(settings);
        } catch (IOException e) {
          page = createWizardPageError("IO-Fehler: " + e.getMessage());
        }
      } else {
        // not initial and last page not known -> log and show error
        logger.error("Unexpected wizard page. The wizard does not know what to do.");
        page = createWizardPageError("Unexpected wizard page. The wizard does not know what to do.");
      }
    }
    return page;
  }

  /**
   * generate code
   * @param gen code generator
   * @param sm state machine
   * @param genTarget target folder
   * @throws IOException
   */
  protected void generateCodeWith(CodeGenerator gen, StateMachine sm, File genTarget) throws IOException {
    gen.generateCodeFor(sm, genTarget);
  }

  /**
   * create finish page
   * @param settings
   * @return wizard page
   */
  protected WizardPage createWizardPageFinish(WizardSettings settings) {
    return new WizardPageFinish((String) settings.get(Wizard.SETT_KEY_GENTARGET));
  }

  /**
   * Create a visible error message, wizard page
   * @param message
   * @return wizard page
   */
  protected WizardPage createWizardPageError(String message) {
    return new WizardPageError(message);
  }

}
