package kcode.statexmi2java.xmireader;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import kcode.statexmi2java.smrepresentation.Behavior;
import kcode.statexmi2java.smrepresentation.State;
import kcode.statexmi2java.smrepresentation.StateMachine;
import kcode.statexmi2java.xmireader.jaxb.*;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * XMI reader parsing XMI file via JAXB (annotated classes)
 */
public class XMIReaderJAXB extends XMIReader {
  protected static final String XMI_TYPE_STATEMACHINE                       = "uml:StateMachine";
  protected static final String XMI_SUBVERTEX_TYPE_PSEUDOSTATE              = "uml:Pseudostate";
  protected static final String XMI_SUBVERTEX_TYPE_STATE                    = "uml:State";
  protected static final Object XMI_SUBVERTEX_TYPE_PSEUDOSTATE_KIND_INITIAL = "initial";
  protected final Logger        logger                                      = LoggerFactory
                                                                                .getLogger(XMIReaderJAXB.class);

  @Override
  public void initialize(String xmiFilePath) {
    super.initialize(xmiFilePath);
    parseXmi();
  }

  protected String xmiVersion;

  private void parseXmi() {
    // new xmi-parse, thus purge statmachine list
    this.stateMachines.clear();
    try {
      final JAXBContext context = JAXBContext.newInstance(XMI.class);
      final Unmarshaller um = context.createUnmarshaller();

      // unmarshall XMI file to object
      final XMI xmi = (XMI) um.unmarshal(new File(this.xmiFilePath));
      // read and set XMI version
      this.xmiVersion = xmi.getVersion();

      // get members and check for statemachines
      Model model = xmi.getModel();
      List<OwnedMember> members = model.getOwnedMembers();
      for (OwnedMember member : members) {
        if (member.getType() != null && member.getType().equals(XMI_TYPE_STATEMACHINE)) {
          logger.debug("Model-OwnedMember-StateMachine found");
          parseXmiOwnedMemberStateMachine(member);
        }
      }
    } catch (JAXBException e) {
      e.printStackTrace();
    }
  }

  /**
   * parse the XMI owned members
   * @param member
   */
  private void parseXmiOwnedMemberStateMachine(OwnedMember member) {
    final List<Region> regions = member.getRegions();
    for (Region region : regions) {

      final List<VertexState> states = new ArrayList<VertexState>();
      final List<VertexState> initialStates = new ArrayList<VertexState>();
      final List<Transition> transitions = new ArrayList<Transition>();

      List<VertexState> vertices = region.getSubvertexes();
      for (VertexState vertex : vertices) {
        if (vertex.getType().equals(XMI_SUBVERTEX_TYPE_PSEUDOSTATE)
            || vertex.getType().equals(XMI_SUBVERTEX_TYPE_STATE)) {
          states.add(vertex);
        }
        if (vertex.getType().equals(XMI_SUBVERTEX_TYPE_PSEUDOSTATE)
            && vertex.getKind().equals(XMI_SUBVERTEX_TYPE_PSEUDOSTATE_KIND_INITIAL)) {
          initialStates.add(vertex);
        }
      }

      List<Transition> xmiTransitions = region.getTransitions();
      for (Transition xmiTransition : xmiTransitions) {
        transitions.add(xmiTransition);
      }

      // now that we have all states and transitions, map them to representation objects
      for (VertexState initialStateVertex : initialStates) {
        State initialState = getStateFromVertex(initialStateVertex);
        StateMachine sm = new StateMachine();
        sm.states.add(initialState);
        resolveStateMachine(sm, initialStateVertex, states, transitions);
        this.stateMachines.add(sm);
      }
    }
  }

  private List<Transition>        mappedTransitions = new ArrayList<Transition>();
  private Map<VertexState, State> mappedStates      = new HashMap<VertexState, State>();

  /**
   * parse state machine
   * @param stateMachine
   * @param stateVertex
   * @param stateVertices
   * @param xmiTransitions
   */
  private void resolveStateMachine(final StateMachine stateMachine, final VertexState stateVertex,
      final List<VertexState> stateVertices, final List<Transition> xmiTransitions) {
    for (Transition xmiTransition : xmiTransitions) {
      Transition xmiTransitionSource = xmiTransition;
      if (xmiTransitionSource.getSource().equals(stateVertex)) {
        if (!mappedTransitions.contains(xmiTransitionSource)) {
          kcode.statexmi2java.smrepresentation.Transition transition;
          transition = getTransitionFromXmiTransition(xmiTransition);
          transition.from = getStateFromVertex(stateVertex);

          VertexState xmiStateTarget = xmiTransition.getTarget();
          State mappedStateTarget;
          if ((mappedStateTarget = mappedStates.get(xmiStateTarget)) == null) {
            mappedStateTarget = getStateFromVertex(xmiStateTarget);
            mappedStates.put(xmiStateTarget, mappedStateTarget);
          }

          if (mappedStateTarget != null) {
            transition.to = mappedStateTarget;
            stateMachine.states.add(mappedStateTarget);
          }
          mappedTransitions.add(xmiTransitionSource);
          resolveStateMachine(stateMachine, xmiTransition.getTarget(), stateVertices, xmiTransitions);
          stateMachine.transitions.add(transition);
        }
      }
    }
  }

  /**
   * get representational state from XMI state (VertexState)
   * @param xmiSubvertex
   * @return
   */
  private State getStateFromVertex(VertexState xmiSubvertex) {
    if (xmiSubvertex == null) {
      return null;
    }
    // is a state subvertex -> create state representation
    State state = new State();
    // if name is set in XMI set it in representation
    if (!xmiSubvertex.getName().isEmpty()) {
      state.name = xmiSubvertex.getName();
    }
    // isInitial?
    if (xmiSubvertex.getKind() != null && xmiSubvertex.getKind().equals("initial")) {
      state.isInitial = true;
    }
    // isFinal?
    if (xmiSubvertex.getKind() != null && xmiSubvertex.getKind().equals("final")) {
      state.isFinal = true;
    }
    // entry, do, exit activities
    kcode.statexmi2java.xmireader.jaxb.Behavior xmiBehEn = xmiSubvertex.getEntryAction();
    if (xmiBehEn != null) {
      state.entry = new Behavior();
      state.entry.name = xmiBehEn.getName();
      state.entry.body = xmiBehEn.getBody();
    }
    kcode.statexmi2java.xmireader.jaxb.Behavior xmiBehDo = xmiSubvertex.getDoAction();
    if (xmiBehDo != null) {
      state.doActivity = new Behavior();
      state.doActivity.name = xmiBehDo.getName();
      state.doActivity.body = xmiBehDo.getBody();
    }
    kcode.statexmi2java.xmireader.jaxb.Behavior xmiBehEx = xmiSubvertex.getExitAction();
    if (xmiBehEx != null) {
      state.exit = new Behavior();
      state.exit.name = xmiBehEx.getName();
      state.exit.body = xmiBehEx.getBody();
    }
    return state;
  }

  /**
   * get mapped, representational transition from XMI transition
   * guards, effect, triggers are mapped, source and target state not
   * @param xmiTransition the XMI transition
   * @return representational transition
   */
  private kcode.statexmi2java.smrepresentation.Transition getTransitionFromXmiTransition(Transition xmiTransition) {

    // parse transition representation
    kcode.statexmi2java.smrepresentation.Transition transition = new kcode.statexmi2java.smrepresentation.Transition();

    // get and set guard condition
    if (xmiTransition.getGuard() != null) {
      Constraint xmiGuard = xmiTransition.getGuard();
      if (xmiGuard != null) {
        Specification spec = xmiGuard.getSpecification();
        if (spec != null) {
          String xmiSpecBody = spec.getBody();
          if (xmiSpecBody != null) {
            transition.guard = xmiSpecBody;
          }
        }
      }
    }

    // get and set triggers
    List<Trigger> xmiTriggers = xmiTransition.getTriggers();
    for (Trigger xmiTrigger : xmiTriggers) {
      // check for valid idref reference in xmi (xmi may include trigger tag with idref pointing to non-existant tag
      if (xmiTrigger.getName() != null) {
        transition.triggers.add(xmiTrigger.getName());
      }
    }

    // transition effects
    kcode.statexmi2java.xmireader.jaxb.Behavior xmiEffect = xmiTransition.getEffect();
    if (xmiEffect != null) {
      Behavior beh = new Behavior();
      beh.name = xmiEffect.getName();
      beh.body = xmiEffect.getBody();
      transition.effect = beh;
    }
    transition.name = xmiTransition.getName();

    return transition;
  }

  @Override
  public String getXMIVersion() {
    return this.xmiVersion;
  }

  @Override
  public int getNumberOfStateDiagrams() {
    return (this.stateMachines != null ? this.stateMachines.size() : 0);
  }

  @Override
  public List<StateMachine> getStateMachines() {
    return stateMachines;
  }
}
