package codetemplates;

import java.util.ArrayList;
import java.util.List;

/**
 * @param <S> StateIDs enum, identifying states
 * @param <T> Triggers enum, representing triggers
 */
public class StateTransition<S extends Enum<S>, T extends Enum<T>> {
  protected List<T> triggers = new ArrayList<T>();
  protected ObjectState<S, T> source;
  protected ObjectState<S, T> target;

  public StateTransition(ObjectState<S, T> source, ObjectState<S, T> target, T trigger) {
    this.source = source;
    source.addTransition(this);
    this.target = target;
    target.addTransition(this);
    this.triggers.add(trigger);
  }

  /**
   * try to trigger a transition
   * @param trigger trigger, triggering the transition
   * @param context a context object
   * @return new state object on success, or null on failure (not triggerable with this trigger, guard denying)
   */
  public ObjectState<S, T> triggerTransitionBy(T trigger, Object context) {
    boolean isTriggerable = isTriggerableBy(trigger, context);
    ObjectState<S, T> os = null;
    if (isTriggerable) {
      executeEffect(context);
      os = doTransition(context);
    }
    return os;
  }

  /**
   * get source state the transition starts from
   * @return source state the transition starts from
   */
  public ObjectState<S, T> getSource() {
    return source;
  }

  /**
   * get target state the transition transitions to
   * @return target state the transition transitions to
   */
  public ObjectState<S, T> getTarget() {
    return target;
  }

  /**
   * check if this transition is triggerable by that trigger
   * also checks for guard
   * @param trigger trigger to check against
   * @param context context object
   * @return true: triggerable, false: not triggerable
   */
  public boolean isTriggerableBy(T trigger, Object context) {
    return isTriggerableBy(trigger, context, true);
  }

  /**
   * check if this transition is triggerable by that trigger
   * @param trigger trigger to check against
   * @param context context object
   * @param checkGuard also check for guard
   * @return true: triggerable, false: not triggerable
   */
  public boolean isTriggerableBy(T trigger, Object context, boolean checkGuard) {
    return triggers.contains(trigger) && isGuardConditionMet(context);
  }

  /**
   * check if guard condition is met
   * @param context context object
   * @return true: guard met, false: not met
   */
  protected boolean isGuardConditionMet(Object context) {
    // default return false -> forces to implement after code generation and prevents default ambiguity where
    // (complementing) guards are supposed to be implemented
    return false;
  }

  /**
   * execute the transitions effect
   * @param context context object
   */
  protected void executeEffect(Object context) {
    // no effect by default – may be overridden by inheriting and concrete classes
  }

  /**
   * do the transition
   * @param context context object
   * @return new state object
   */
  protected ObjectState<S, T> doTransition(Object context) {
    source.onStateExit(context);
    target.onStateEntry(context);
    ObjectState<S, T> os = target;
    return os;
  }
}
