package kcode.statexmi2java.xmireader;

import static org.junit.Assert.*;

import java.io.File;

import kcode.statexmi2java.smrepresentation.*;
import kcode.statexmi2java.xmireader.XMIReaderJAXB;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;

public class XMIReaderJAXBTest {
  static final String PATH                           = System.getProperty("user.dir") + File.separator + "docs"
                                                         + File.separator + "xmi-examples-vp";
  static final String XMIFILE_BLANK                  = "f01-blank.xmi";
  static final String XMIFILE_SM_EMPTY               = "f02-sm.xmi";
  static final String XMIFILE_INITIAL_FINAL          = "f03-initial-final.xmi";
  static final String XMIFILE_TRANSITION             = "f04-transition.xmi";
  static final String XMIFILE_STATE_TRANSITIONS      = "f05-state-transitions.xmi";
  static final String XMIFILE_ST_NAMED_GUARD         = "f06-state-transitions-named-guard.xmi";
  static final String XMIFILE_ST_NG_TRIGGER_DOACTION = "f07-state-transitions-named-guard-trigger-doaction.xmi";
  static final String XMIFILE_ST_NG_ACTION_EFFECT    = "f08-state-transitions-named-guard-trigger-doaction-effect.xmi";
  static final String XMIFILE_ST_NG_AE_EE            = "f09-state-transitions-named-guard-trigger-doaction-effect-entryexit.xmi";
  static final String XMIFILE_2SMS                   = "2sms.xmi";

  XMIReaderJAXB       testee;

  @Before
  public void setUp() throws Exception {
    testee = new XMIReaderJAXB();
  }

  @After
  public void tearDown() throws Exception {
    testee = null;
  }

  @Test
  public void testGetXMIVersion() {
    testee.initialize(PATH + File.separator + XMIFILE_BLANK);
    assertEquals("2.1", testee.getXMIVersion());
  }

  @Test
  public void testGetNoSM() {
    testee.initialize(PATH + File.separator + XMIFILE_BLANK);
    assertEquals(0, testee.getStateMachines().size());
    assertTrue(testee.getStateMachines().isEmpty());
  }

  @Test
  public void testGetEmptySM() {
    // ~~~ VP will drop empty statemachine from model and only add it do diagram repr. in XMI
    // thus, as no SM is found in the model, null (^= no SM) is expected
    testee.initialize(PATH + File.separator + XMIFILE_SM_EMPTY);
    assertEquals(0, testee.getStateMachines().size());
    assertTrue(testee.getStateMachines().isEmpty());
  }

  @Test
  public void testGetSMWithInitialFinal() {
    testee.initialize(PATH + File.separator + XMIFILE_INITIAL_FINAL);

    StateMachine expected = new StateMachine();
    State s = new State();
    s.name = "Initial";
    s.isInitial = true;
    expected.states.add(s);
    s = new State();
    s.name = "FinalState";
    expected.states.add(s);

    assertEquals(expected, testee.getStateMachines().get(0));
  }

  @Test
  public void testGetSMWithTransition() {
    testee.initialize(PATH + File.separator + XMIFILE_TRANSITION);

    StateMachine exp = new StateMachine();
    State sI = new State();
    sI.name = "Initial";
    sI.isInitial = true;
    exp.states.add(sI);
    State sF = new State();
    sF.name = "FinalState";
    sF.isFinal = true;
    exp.states.add(sF);
    Transition t = new Transition();
    t.name = "";
    t.from = sI;
    t.to = sF;
    t.guard = "";
    exp.transitions.add(t);

    assertEquals(exp, testee.getStateMachines().get(0));
  }

  @Test
  public void testGetSMWithTransitions() {
    testee.initialize(PATH + File.separator + XMIFILE_STATE_TRANSITIONS);

    StateMachine exp = new StateMachine();
    State sI = new State();
    sI.name = "Initial";
    sI.isInitial = true;
    exp.states.add(sI);
    State sF = new State();
    sF.name = "FinalState";
    sF.isFinal = true;
    exp.states.add(sF);
    State sM = new State();
    sM.name = "MyState";
    exp.states.add(sM);
    Transition tS = new Transition();
    tS.name = "";
    tS.from = sI;
    tS.to = sM;
    tS.guard = "";
    exp.transitions.add(tS);
    Transition tT = new Transition();
    tT.name = "";
    tT.from = sM;
    tT.to = sF;
    tT.guard = "";
    exp.transitions.add(tT);

    assertEquals(exp, testee.getStateMachines().get(0));
  }

  @Test
  public void testGetSMWithTNamedGuard() {
    testee.initialize(PATH + File.separator + XMIFILE_ST_NAMED_GUARD);

    StateMachine exp = new StateMachine();
    State sI = new State();
    sI.name = "Initial";
    sI.isInitial = true;
    exp.states.add(sI);
    State sF = new State();
    sF.name = "FinalState";
    sF.isFinal = true;
    exp.states.add(sF);
    State sM = new State();
    sM.name = "MyState";
    exp.states.add(sM);
    Transition tS = new Transition();
    tS.name = "become";
    tS.from = sI;
    tS.to = sM;
    tS.guard = "myAlways";
    exp.transitions.add(tS);
    Transition tT = new Transition();
    tT.name = "begone";
    tT.from = sM;
    tT.to = sF;
    tT.guard = "myAlways";
    exp.transitions.add(tT);

    assertEquals(exp, testee.getStateMachines().get(0));
  }

  @Test
  public void testGetSMWithTNGTriggerDoaction() {
    testee.initialize(PATH + File.separator + XMIFILE_ST_NG_TRIGGER_DOACTION);

    StateMachine exp = new StateMachine();
    State sI = new State();
    sI.name = "Initial";
    sI.isInitial = true;
    exp.states.add(sI);
    State sF = new State();
    sF.name = "FinalState";
    sF.isFinal = true;
    exp.states.add(sF);
    State sM = new State();
    sM.name = "MyState";
    Behavior doA = new Behavior();
    doA.name = "myDoActivity";
    doA.body = "myDoingSomething";
    sM.doActivity = doA;
    exp.states.add(sM);
    Transition transInit = new Transition();
    transInit.name = "become";
    transInit.from = sI;
    transInit.to = sM;
    transInit.guard = "myAlways";
    transInit.triggers.add("myTrigger");
    exp.transitions.add(transInit);
    Transition transFin = new Transition();
    transFin.name = "begone";
    transFin.from = sM;
    transFin.to = sF;
    transFin.guard = "myAlways";
    exp.transitions.add(transFin);

    assertEquals(exp, testee.getStateMachines().get(0));
  }

  @Test
  public void testGetSMWithTNGTDActionEffect() {
    testee.initialize(PATH + File.separator + XMIFILE_ST_NG_ACTION_EFFECT);

    StateMachine exp = new StateMachine();
    State sI = new State();
    sI.name = "Initial";
    sI.isInitial = true;
    exp.states.add(sI);
    State sF = new State();
    sF.name = "FinalState";
    sF.isFinal = true;
    exp.states.add(sF);
    State sM = new State();
    sM.name = "MyState";
    Behavior doA = new Behavior();
    doA.name = "myDoActivity";
    doA.body = "myDoingSomething";
    sM.doActivity = doA;
    exp.states.add(sM);
    Transition tS = new Transition();
    tS.name = "become";
    tS.from = sI;
    tS.to = sM;
    tS.guard = "myAlways";
    tS.triggers.add("myTrigger");
    exp.transitions.add(tS);
    Transition tT = new Transition();
    tT.name = "begone";
    tT.from = sM;
    tT.to = sF;
    tT.guard = "myAlways";
    // effect
    Behavior behEff = new Behavior();
    behEff.name = "myEffect";
    behEff.body = "myDoMyEffect";
    tT.effect = behEff;
    exp.transitions.add(tT);

    assertEquals(exp, testee.getStateMachines().get(0));
  }

  @Test
  public void testGetSMWithTNGTDAE() {
    testee.initialize(PATH + File.separator + XMIFILE_ST_NG_AE_EE);

    StateMachine exp = new StateMachine();
    State sI = new State();
    sI.name = "Initial";
    sI.isInitial = true;
    exp.states.add(sI);
    State sF = new State();
    sF.name = "FinalState";
    sF.isFinal = true;
    exp.states.add(sF);
    State sM = new State();
    sM.name = "MyState";
    Behavior doA = new Behavior();
    doA.name = "myDoActivity";
    doA.body = "myDoingSomething";
    sM.doActivity = doA;
    // Entry
    Behavior en = new Behavior();
    en.name = "myEntry";
    en.body = "myEntryBody";
    sM.entry = en;
    // Exit
    Behavior ex = new Behavior();
    ex.name = "myExit";
    ex.body = "myExitBody";
    sM.exit = ex;
    exp.states.add(sM);
    Transition tS = new Transition();
    tS.name = "become";
    tS.from = sI;
    tS.to = sM;
    tS.guard = "myAlways";
    exp.transitions.add(tS);
    Transition tT = new Transition();
    tT.name = "begone";
    tT.from = sM;
    tT.to = sF;
    tT.guard = "myAlways";
    // effect
    Behavior behEff = new Behavior();
    behEff.name = "myEffect";
    behEff.body = "myDoMyEffect";
    tT.effect = behEff;
    exp.transitions.add(tT);

    assertEquals(exp, testee.getStateMachines().get(0));
  }

  @Test
  public void testGetSM2StateMachines() {
    testee.initialize(PATH + File.separator + XMIFILE_2SMS);

    StateMachine exp1 = new StateMachine();
    State sI1 = new State();
    sI1.name = "Initial";
    sI1.isInitial = true;
    exp1.states.add(sI1);
    State s1 = new State();
    s1.name = "MyState1";
    exp1.states.add(s1);
    Transition tS1 = new Transition();
    tS1.from = sI1;
    tS1.to = s1;
    tS1.name = "";
    tS1.guard = "";
    exp1.transitions.add(tS1);

    StateMachine exp2 = new StateMachine();
    State sI2 = new State();
    sI2.name = "Initial2";
    sI2.isInitial = true;
    exp2.states.add(sI2);
    State s2 = new State();
    s2.name = "MyState2";
    exp2.states.add(s2);
    Transition tS2 = new Transition();
    tS2.from = sI2;
    tS2.to = s2;
    tS2.name = "";
    tS2.guard = "";
    exp2.transitions.add(tS2);

    assertEquals(2, testee.getStateMachines().size());
    assertEquals(exp1, testee.getStateMachines().get(0));
    assertEquals(exp2, testee.getStateMachines().get(1));
  }
}
